import asyncio
import copy
import time
from asyncio.coroutines import iscoroutinefunction
from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, Callable, Generic, Iterator, TypeVar

from aiohubspace.errors import DeviceNotFound, ExceededMaximumRetries

from .. import v1_const
from ..device import HubspaceDevice, get_hs_device
from ..models.resource import ResourceTypes
from .event import EventCallBackType, EventType, HubspaceEvent

if TYPE_CHECKING:
    from .. import HubspaceBridgeV1


EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
]

ID_FILTER_ALL = "*"

HubspaceResource = TypeVar("HubspaceResource")


class BaseResourcesController(Generic[HubspaceResource]):
    """Base Controller for Hubspace devices"""

    ITEM_TYPE_ID: ResourceTypes | None = None
    ITEM_TYPES: list[ResourceTypes] | None = None
    ITEM_CLS = None
    ITEM_MAPPING: dict = {}

    def __init__(self, bridge: "HubspaceBridgeV1") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._items: dict[str, HubspaceResource] = {}
        self._logger = bridge.logger.getChild(self.ITEM_CLS.__name__)
        self._subscribers: dict[str, EventSubscriptionType] = {ID_FILTER_ALL: []}
        self._initialized: bool = False
        self._item_values = [x.value for x in self.ITEM_TYPES]

    def __getitem__(self, device_id: str) -> HubspaceResource:
        """Get item by device_id."""
        return self._items[device_id]

    def __iter__(self) -> Iterator[HubspaceResource]:
        """Iterate items."""
        return iter(self._items.values())

    def __contains__(self, device_id: str) -> bool:
        """Return bool if device_id is in items."""
        return device_id in self._items

    @property
    def items(self) -> list[HubspaceResource]:
        """Return all items for this resource."""
        return list(self._items.values())

    @property
    def initialized(self) -> bool:
        return self._initialized

    async def _handle_event(
        self, evt_type: EventType, evt_data: HubspaceEvent | None
    ) -> None:
        """Handle incoming event for this resource"""
        if evt_data is None:
            return
        item_id = evt_data.get("device_id", None)
        if evt_type == EventType.RESOURCE_ADDED:
            cur_item = await self.initialize_elem(evt_data["device"])
        elif evt_type == EventType.RESOURCE_DELETED:
            cur_item = self._items.pop(item_id, evt_data)
        elif evt_type == EventType.RESOURCE_UPDATED:
            # existing item updated
            try:
                cur_item = self.get_device(item_id)
            except DeviceNotFound:
                cur_item = None
            if cur_item is None:
                return
            # @TODO - Check for changes rather than issuing a full update for everything each time
            await self.update_elem(evt_data["device"])
        else:
            # Skip all other events
            return
        subscribers = (
            self._subscribers.get(item_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for callback, event_filter in subscribers:
            if event_filter is not None and evt_type not in event_filter:
                continue
            # dispatch the full resource object to the callback
            if iscoroutinefunction(callback):
                asyncio.create_task(callback(evt_type, cur_item))
            else:
                callback(evt_type, cur_item)

    async def initialize(self, initial_data: list[dict]) -> None:
        """Initialize controller by fetching all items for this resource type from bridge."""
        if self._initialized:
            return
        valid_devices: list[HubspaceDevice] = []
        try:
            valid_devices = self.get_filtered_devices(initial_data)
        except AttributeError:
            for element in initial_data:
                if element["typeId"] != self.ITEM_TYPE_ID.value:
                    self._logger.debug(
                        "TypeID [%s] does not match %s",
                        element["typeId"],
                        self.ITEM_TYPE_ID.value,
                    )
                    continue
                device = get_hs_device(element)
                if device.device_class not in self._item_values:
                    self._logger.debug(
                        "Device Class [%s] is not contained in %s",
                        device.device_class,
                        self._item_values,
                    )
                    continue
                valid_devices.append(device)

        for item in valid_devices:
            await self._handle_event(
                EventType.RESOURCE_ADDED,
                HubspaceEvent(
                    type=EventType.RESOURCE_ADDED,
                    device_id=item.device_id,
                    device=item,
                ),
            )
        # subscribe to item updates
        res_filter = [x.value for x in self.ITEM_TYPES]
        self._bridge.events.subscribe(
            self._handle_event,
            resource_filter=tuple(res_filter),
        )
        self._initialized = True

    async def initialize_elem(self, element: HubspaceDevice) -> None:
        raise NotImplementedError("Class should implement initialize_elem")

    async def update_elem(self, element: HubspaceDevice) -> None:
        raise NotImplementedError("Class should implement update_elem")

    def subscribe(
        self,
        callback: EventCallBackType,
        id_filter: str | tuple[str] | None = None,
        event_filter: EventType | tuple[EventType] | None = None,
    ) -> Callable:
        """
        Subscribe to status changes for this resource type.

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `id_filter` - Optionally provide resource ID(s) to filter events for.
            - `event_filter` - Optionally provide EventType(s) as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, None | list | tuple):
            event_filter = (event_filter,)

        if id_filter is None:
            id_filter = (ID_FILTER_ALL,)
        elif not isinstance(id_filter, list | tuple):
            id_filter = (id_filter,)

        subscription = (callback, event_filter)

        for id_key in id_filter:
            if id_key not in self._subscribers:
                self._subscribers[id_key] = []
            self._subscribers[id_key].append(subscription)

        # unsubscribe logic
        def unsubscribe():
            for id_key in id_filter:
                if id_key not in self._subscribers:
                    continue
                self._subscribers[id_key].remove(subscription)

        return unsubscribe

    async def update(
        self,
        device_id: str,
        obj_in: Generic[HubspaceResource],
    ) -> None:
        """Update Hubspace with the new data

        :param device_id: Hubspace Device ID
        :param obj_in: Hubspace Resource elements to change
        """
        cur_item = self._items.get(device_id)
        if cur_item is None:
            return
        hs_states = dataclass_to_hs(cur_item, obj_in, self.ITEM_MAPPING)
        if not hs_states:
            self._logger.debug("No states to send. Skipping")
            return
        # Make a clone if the update fails
        fallback = copy.deepcopy(cur_item)
        # Update the state of the item to match the new states
        update_dataclass(cur_item, obj_in)
        # If the update fails, restore the old states
        fallback_required: bool = False
        # @TODO - Implement bluetooth logic for update
        if True:
            url = v1_const.HUBSPACE_DEVICE_STATE.format(
                self._bridge.account_id, str(device_id)
            )
            headers = {
                "host": v1_const.HUBSPACE_DATA_HOST,
                "content-type": "application/json; charset=utf-8",
            }
            payload = {"metadeviceId": str(device_id), "values": hs_states}
            try:
                res = await self._bridge.request(
                    "put", url, json=payload, headers=headers
                )
            except ExceededMaximumRetries:
                fallback_required = True
            else:
                # Bad states provided
                if res.status == 400:
                    self._logger.warning(
                        "Invalid update provided for %s using %s", device_id, hs_states
                    )
                    fallback_required = True
        if fallback_required:
            self._items[device_id] = fallback

    def get_device(self, device_id) -> HubspaceResource:
        cur_item = self._items.get(device_id)
        if cur_item is None:
            raise DeviceNotFound(device_id)
        return cur_item


def update_dataclass(elem: HubspaceResource, cls: dataclass):
    """Updates the element with the latest changes"""
    for f in fields(cls):
        cur_val = getattr(cls, f.name, None)
        elem_val = getattr(elem, f.name)
        if cur_val is None:
            continue
        # Special processing for dicts
        if isinstance(elem_val, dict):
            cur_val = {getattr(cur_val, "func_instance", None): cur_val}
            getattr(elem, f.name).update(cur_val)
        else:
            setattr(elem, f.name, cur_val)


def dataclass_to_hs(
    elem: HubspaceResource, cls: dataclass, mapping: dict
) -> list[dict]:
    """Convert the current state to be consumed by Hubspace"""
    states = []
    for f in fields(cls):
        cur_val = getattr(cls, f.name, None)
        if cur_val is None:
            continue
        if cur_val == getattr(elem, f.name, None):
            continue
        hs_key = mapping.get(f.name, f.name)
        new_val = cur_val.hs_value
        if not isinstance(new_val, list):
            new_val = [new_val]
        for val in new_val:
            new_state = {
                "functionClass": hs_key,
                "functionInstance": elem.get_instance(hs_key),
                "lastUpdateTime": int(time.time()),
                "value": None,
            }
            if isinstance(val, dict):
                new_state.update(val)
            else:
                new_state["value"] = val
            states.append(new_state)
    return states



from typing import Final

HUBSPACE_DEFAULT_USERAGENT: Final[str] = "Dart/2.15 (dart:io)"
HUBSPACE_ACCOUNT_ID_URL: Final[str] = "https://api2.afero.net/v1/users/me"
HUBSPACE_DEFAULT_ENCODING: Final[str] = "gzip"

HUBSPACE_DATA_URL: Final[str] = "https://api2.afero.net/v1/accounts/{}/metadevices"

HUBSPACE_DEVICE_STATE: Final[str] = (
    "https://api2.afero.net/v1/accounts/{}/metadevices/{}/state"
)
HUBSPACE_DATA_HOST: Final[str] = "semantics2.afero.net"

DEFAULT_HEADERS: Final[dict[str, str]] = {
    "user-agent": HUBSPACE_DEFAULT_USERAGENT,
    "accept-encoding": HUBSPACE_DEFAULT_ENCODING,
}

MAX_RETRIES: Final[int] = 3



"""Controller holding and managing Hubspace resources of type `lock`."""

from ..device import HubspaceDevice
from ..models import features, lock
from ..models.resource import DeviceInformation, ResourceTypes
from .base import BaseResourcesController


class LockController(BaseResourcesController[lock.Lock]):
    """Controller holding and managing Hubspace resources of type `lock`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.LOCK]
    ITEM_CLS = lock.Lock
    ITEM_MAPPING = {"position": "lock-control"}

    async def lock(self, device_id: str) -> None:
        """Engage the lock"""
        await self.set_state(
            device_id, lock_position=features.CurrentPositionEnum.LOCKING
        )

    async def unlock(self, device_id: str) -> None:
        """Disengage the lock"""
        await self.set_state(
            device_id, lock_position=features.CurrentPositionEnum.UNLOCKING
        )

    async def initialize_elem(self, hs_device: HubspaceDevice) -> None:
        """Initialize the element"""
        self._logger.info("Initializing %s", hs_device.id)
        available: bool = False
        current_position: features.CurrentPositionFeature | None = None
        for state in hs_device.states:
            if state.functionClass == "lock-control":
                current_position = features.CurrentPositionFeature(
                    position=features.CurrentPositionEnum(state.value)
                )
            elif state.functionClass == "available":
                available = state.value

        self._items[hs_device.id] = lock.Lock(
            hs_device.functions,
            id=hs_device.id,
            available=available,
            device_information=DeviceInformation(
                device_class=hs_device.device_class,
                default_image=hs_device.default_image,
                default_name=hs_device.default_name,
                manufacturer=hs_device.manufacturerName,
                model=hs_device.model,
                name=hs_device.friendly_name,
                parent_id=hs_device.device_id,
            ),
            position=current_position,
        )

    async def update_elem(self, hs_device: HubspaceDevice) -> None:
        cur_item = self.get_device(hs_device.id)
        for state in hs_device.states:
            if state.functionClass == "lock-control":
                cur_item.position.position = features.CurrentPositionEnum(state.value)
            elif state.functionClass == "available":
                cur_item.available = state.value

    async def set_state(
        self,
        device_id: str,
        lock_position: features.CurrentPositionEnum | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = lock.LockPut()
        if lock_position is not None:
            update_obj.position = features.CurrentPositionFeature(
                position=lock_position
            )
        await self.update(device_id, update_obj)



import json
import os
import pathlib
import time
from urllib.parse import urlencode

import aiohttp
import pytest
import pytest_asyncio

from aiohubspace.v1 import auth

current_path = pathlib.Path(__file__).parent.resolve()


@pytest.fixture(scope="function")
def hs_auth():
    return auth.HubspaceAuth("username", "password")


@pytest_asyncio.fixture(scope="function")
async def aio_sess() -> aiohttp.ClientSession:
    yield aiohttp.ClientSession()


async def build_url(base_url: str, qs: dict[str, str]) -> str:
    return f"{base_url}?{urlencode(qs)}"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "time_offset,is_expired",
    [
        # No token
        (None, True),
        # Expired token
        (-5, True),
        # Non-Expired token
        (5, False),
    ],
)
async def test_is_expired(time_offset, is_expired, hs_auth):
    if time_offset:
        hs_auth._token_data = auth.token_data("token", time.time() + time_offset)
    assert await hs_auth.is_expired == is_expired


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "page_filename,err_msg,expected",
    [
        # Valid
        (
            "auth_webapp_login.html",
            None,
            auth.auth_sess_data("url_sess_code", "url_exec_code", "url_tab_id"),
        ),
        # page is missing expected id
        (
            "auth_webapp_login_missing.html",
            "Unable to parse login page",
            None,
        ),
        # form field is missing expected attribute
        (
            "auth_webapp_login_bad_format.html",
            "Unable to extract login url",
            None,
        ),
        # URL missing expected elements
        (
            "auth_webapp_login_bad_qs.html",
            "Unable to parse login url",
            None,
        ),
    ],
)
async def test_extract_login_data(page_filename, err_msg, expected):
    with open(os.path.join(current_path, "data", page_filename), "r") as f:
        page_data = f.read()
    if expected:
        assert await auth.extract_login_data(page_data) == expected
    else:
        with pytest.raises(auth.InvalidResponse, match=err_msg):
            await auth.extract_login_data(page_data)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "page_filename, gc_exp, response, expected_err",
    [
        # Invalid status code
        (None, None, {"status": 403}, aiohttp.ClientError),
        # Invalid auth provided
        (None, None, {"status": 200}, aiohttp.ClientError),
        # @TODO - Correctly match param escaping
        # Valid auth passed to generate_code
        # ("auth_webapp_login.html", None, {"status": 200}, None),
    ],
)
async def test_webapp_login(
    page_filename,
    gc_exp,
    response,
    expected_err,
    hs_auth,
    mock_aioresponse,
    aio_sess,
    mocker,
):
    if page_filename:
        with open(os.path.join(current_path, "data", page_filename), "r") as f:
            response["body"] = f.read()
    challenge = await hs_auth.generate_challenge_data()
    generate_code = mocker.patch.object(hs_auth, "generate_code")
    params: dict[str, str] = {
        "response_type": "code",
        "client_id": auth.HUBSPACE_DEFAULT_CLIENT_ID,
        "redirect_uri": auth.HUBSPACE_DEFAULT_REDIRECT_URI,
        "code_challenge": challenge.challenge,
        "code_challenge_method": "S256",
        "scope": "openid offline_access",
    }
    url = await build_url(auth.HUBSPACE_OPENID_URL, params)
    mock_aioresponse.post(url, **response)
    if not expected_err:
        await hs_auth.webapp_login(challenge, aio_sess)
        generate_code.assert_called_once_with(*gc_exp)
    else:
        with pytest.raises(expected_err):
            await hs_auth.webapp_login(challenge, aio_sess)
        generate_code.assert_not_called()


@pytest.mark.asyncio
async def test_generate_challenge_data():
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "session_code, execution, tab_id, response, expected_err, expected",
    [
        # Invalid response
        (
            "sess_code",
            "execution",
            "tab_id",
            {"status": 200},
            auth.InvalidAuth,
            None,
        ),
        # Invalid Location
        (
            "sess_code",
            "execution",
            "tab_id",
            {"status": 302, "headers": {"location": "nope"}},
            auth.InvalidResponse,
            None,
        ),
        # Valid location
        (
            "sess_code",
            "execution",
            "tab_id",
            {"status": 302, "headers": {"location": "https://cool.beans?code=beans"}},
            None,
            "beans",
        ),
    ],
)
async def test_generate_code(
    session_code,
    execution,
    tab_id,
    response,
    expected_err,
    expected,
    hs_auth,
    aioresponses,
    aio_sess,
):
    params = {
        "session_code": session_code,
        "execution": execution,
        "client_id": auth.HUBSPACE_DEFAULT_CLIENT_ID,
        "tab_id": tab_id,
    }
    url = await build_url(auth.HUBSPACE_CODE_URL, params)
    aioresponses.post(url, **response)
    if not expected_err:
        assert (
            await hs_auth.generate_code(session_code, execution, tab_id, aio_sess)
            == expected
        )
    else:
        with pytest.raises(expected_err):
            await hs_auth.generate_code(session_code, execution, tab_id, aio_sess)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "code,response,expected,err",
    [
        # Invalid refresh token
        ("code", {"status": 403}, None, aiohttp.ClientError),
        # Incorrect format
        (
            "code",
            {"status": 200, "body": json.dumps({"refresh_token2": "cool_beans"})},
            None,
            auth.InvalidResponse,
        ),
        # Valid refresh token
        (
            "code",
            {"status": 200, "body": json.dumps({"refresh_token": "cool_beans"})},
            "cool_beans",
            None,
        ),
    ],
)
async def test_generate_refresh_token(
    code, response, expected, err, hs_auth, aioresponses, aio_sess
):
    challenge = await hs_auth.generate_challenge_data()
    aioresponses.post(auth.HUBSPACE_TOKEN_URL, **response)
    if expected:
        assert expected == await hs_auth.generate_refresh_token(
            code, challenge, aio_sess
        )
    else:
        with pytest.raises(err):
            await hs_auth.generate_refresh_token(code, challenge, aio_sess)
    aioresponses.assert_called_once()
    call_args = list(aioresponses.requests.values())[0][0]
    assert call_args.kwargs["headers"] == auth.HUBSPACE_TOKEN_HEADERS
    assert call_args.kwargs["data"] == {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": auth.HUBSPACE_DEFAULT_REDIRECT_URI,
        "code_verifier": challenge.verifier,
        "client_id": auth.HUBSPACE_DEFAULT_CLIENT_ID,
    }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "refresh_token,response,expected,err",
    [
        # Invalid status
        ("code", {"status": 403}, None, aiohttp.ClientError),
        # bad response
        (
            "code",
            {"status": 200, "body": json.dumps({"id_token2": "cool_beans"})},
            None,
            auth.InvalidResponse,
        ),
        # valid response
        (
            "code",
            {"status": 200, "body": json.dumps({"id_token": "cool_beans"})},
            "cool_beans",
            None,
        ),
    ],
)
async def test_generate_token(
    refresh_token, response, expected, err, aioresponses, aio_sess
):
    aioresponses.post(auth.HUBSPACE_TOKEN_URL, **response)
    if expected:
        assert expected == (await auth.generate_token(aio_sess, refresh_token)).token
    else:
        with pytest.raises(err):
            await auth.generate_token(aio_sess, refresh_token)
    aioresponses.assert_called_once()
    call_args = list(aioresponses.requests.values())[0][0]
    assert call_args.kwargs["headers"] == auth.HUBSPACE_TOKEN_HEADERS
    assert call_args.kwargs["data"] == {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "scope": "openid email offline_access profile",
        "client_id": "hubspace_android",
    }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "webapp_login_return, generate_refresh_token_return",
    [
        ("cool", "beans"),
    ],
)
async def test_perform_initial_login(
    webapp_login_return, generate_refresh_token_return, hs_auth, aio_sess, mocker
):
    mocker.patch.object(hs_auth, "webapp_login", return_value=webapp_login_return)
    mocker.patch.object(
        hs_auth, "generate_refresh_token", return_value=generate_refresh_token_return
    )
    assert (
        await hs_auth.perform_initial_login(aio_sess) == generate_refresh_token_return
    )


# @TODO - Implement this test
# async def test_token():
#     pass



from contextlib import suppress
from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes


@dataclass
class Light[HubspaceResource]:
    """Representation of a Hubspace Light"""

    id: str  # ID used when interacting with Hubspace
    available: bool

    on: features.OnFeature
    color: features.ColorFeature
    color_mode: features.ColorModeFeature
    color_temperature: features.ColorTemperatureFeature
    dimming: features.DimmingFeature
    effect: features.EffectFeature

    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)

    type: ResourceTypes = ResourceTypes.LIGHT

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            with suppress(KeyError):
                if function["functionInstance"]:
                    instances[function["functionClass"]] = function["functionInstance"]
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)

    @property
    def supports_color(self) -> bool:
        """Return if this light supports color control."""
        return self.color is not None

    @property
    def supports_color_temperature(self) -> bool:
        """Return if this light supports color_temperature control."""
        return self.color_temperature is not None

    @property
    def supports_dimming(self) -> bool:
        """Return if this light supports brightness control."""
        return self.dimming is not None

    @property
    def supports_effects(self) -> bool:
        """Return if this light supports brightness control."""
        return self.effect is not None

    @property
    def supports_on(self):
        return self.on is not None

    @property
    def is_on(self) -> bool:
        """Return bool if light is currently powered on."""
        if self.on is not None:
            return self.on.on
        return False

    @property
    def brightness(self) -> float:
        """Return current brightness of light."""
        if self.dimming is not None:
            return self.dimming.brightness
        return 100.0 if self.is_on else 0.0


@dataclass
class LightPut[HubspaceResource]:
    """States that can be updated for a light"""

    on: features.OnFeature | None = None
    color: features.ColorFeature | None = None
    color_mode: features.ColorModeFeature | None = None
    color_temperature: features.ColorTemperatureFeature | None = None
    dimming: features.DimmingFeature | None = None
    effect: features.EffectFeature | None = None



__all__ = ["HubspaceBridgeV1", "HubspaceError", "InvalidAuth", "InvalidResponse"]


from importlib.metadata import PackageNotFoundError, version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "aiohubspace"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from .errors import HubspaceError, InvalidAuth, InvalidResponse
from .v1 import HubspaceBridgeV1



import pytest
from aioresponses import aioresponses

from aiohubspace.v1 import HubspaceBridgeV1


@pytest.fixture
def mocked_bridge(mocker):
    mocked_bridge = mocker.Mock(HubspaceBridgeV1, autospec=True)("user", "passwd")
    mocker.patch.object(mocked_bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(mocked_bridge, "request", side_effect=mocker.AsyncMock())
    yield mocked_bridge


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m



from dataclasses import dataclass, field

from .resource import DeviceInformation, ResourceTypes
from .sensor import HubspaceSensor


@dataclass
class Device:
    """Representation of a Hubspace parent item"""

    id: str  # ID used when interacting with Hubspace
    available: bool

    sensors: dict[str, HubspaceSensor] = field(default_factory=dict)
    binary_sensors: dict[str, HubspaceSensor] = field(default_factory=dict)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)

    type: ResourceTypes = ResourceTypes.PARENT_DEVICE



"""Controller that holds top-level devices"""

import re
from typing import Any

from ..device import HubspaceDevice, HubspaceState, get_hs_device
from ..models import device, sensor
from ..models.resource import DeviceInformation, ResourceTypes
from .base import BaseResourcesController

unit_extractor = re.compile(r"(\d*)(\D*)")

SENSOR_TO_UNIT: dict[str, str] = {
    "power": "W",
    "watts": "W",
    "wifi-rssi": "dB",
}


class DeviceController(BaseResourcesController[device.Device]):
    """Controller that identifies top-level components."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = []
    ITEM_CLS = device.Device

    async def initialize_elem(self, hs_device: HubspaceDevice) -> None:
        """Initialize the element"""
        self._logger.info("Initializing %s", hs_device.id)
        available: bool = False
        sensors: dict[str, sensor.HubspaceSensor] = {}
        binary_sensors: dict[str, sensor.HubspaceSensor] = {}
        for state in hs_device.states:
            if state.functionClass == "available":
                available = state.value
            elif state.functionClass in sensor.MAPPED_SENSORS:
                value, unit = split_sensor_data(state)
                sensors[state.functionClass] = sensor.HubspaceSensor(
                    id=state.functionClass,
                    owner=hs_device.device_id,
                    value=value,
                    unit=unit,
                )
            elif state.functionClass in sensor.BINARY_SENSORS:
                value, unit = split_sensor_data(state)
                key = f"{state.functionClass}|{state.functionInstance}"
                binary_sensors[key] = sensor.HubspaceSensor(
                    id=key,
                    owner=hs_device.device_id,
                    value=value,
                    unit=unit,
                    instance=state.functionInstance,
                )

        self._items[hs_device.id] = device.Device(
            id=hs_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=hs_device.device_class,
                default_image=hs_device.default_image,
                default_name=hs_device.default_name,
                manufacturer=hs_device.manufacturerName,
                model=hs_device.model,
                name=hs_device.friendly_name,
                parent_id=hs_device.device_id,
            ),
        )

    def get_filtered_devices(self, initial_data: list[dict]) -> list[dict]:
        """Find parent devices"""
        parents: dict = {}
        for element in initial_data:
            if element["typeId"] != self.ITEM_TYPE_ID.value:
                self._logger.debug(
                    "TypeID [%s] does not match %s",
                    element["typeId"],
                    self.ITEM_TYPE_ID.value,
                )
                continue
            device: HubspaceDevice = get_hs_device(element)
            if device.device_id not in parents or device.children:
                parents[device.device_id] = device
        return list(parents.values())

    async def update_elem(self, hs_device: HubspaceDevice) -> None:
        cur_item = self.get_device(hs_device.id)
        for state in hs_device.states:
            if state.functionClass == "available":
                cur_item.available = state.value
            elif state.functionClass in sensor.MAPPED_SENSORS:
                cur_item.sensors[state.functionClass].value = state.value


def split_sensor_data(state: HubspaceState) -> tuple[Any, str | None]:
    if isinstance(state.value, str):
        if match := unit_extractor.match(state.value):
            if (val := match.group(1)) and (unit := match.group(2)):
                return val, unit
    return state.value, SENSOR_TO_UNIT.get(state.functionClass, None)



"""Controls Hubspace devices on v1 API"""

__all__ = ["HubspaceBridgeV1", "InvalidAuth", "InvalidResponse", "HubspaceError"]

import asyncio
import copy
import logging
from contextlib import asynccontextmanager
from types import TracebackType
from typing import Any, Callable, Generator, Optional

import aiohttp
from aiohttp import web_exceptions

from ..errors import ExceededMaximumRetries, HubspaceError, InvalidAuth, InvalidResponse
from . import v1_const
from .auth import HubspaceAuth
from .controllers.device import DeviceController
from .controllers.event import EventCallBackType, EventStream
from .controllers.fan import FanController
from .controllers.light import LightController
from .controllers.lock import LockController
from .controllers.switch import SwitchController
from .controllers.valve import ValveController


class HubspaceBridgeV1:
    """Controls Hubspace devices on v1 API"""

    _web_session: Optional[aiohttp.ClientSession] = None

    def __init__(
        self,
        username: str,
        password: str,
        session: Optional[aiohttp.ClientSession] = None,
        polling_interval: int = 30,
    ):
        self._close_session: bool = session is None
        self._web_session: aiohttp.ClientSession = session
        self._account_id: Optional[str] = None
        self._auth = HubspaceAuth(username, password)
        self.logger = logging.getLogger(f"{__package__}[{username}]")
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(logging.DEBUG)
        self._known_devs: set[str] = set()
        self._known_dev_classes = {}
        # Data Updater
        self._events: EventStream = EventStream(self, polling_interval)
        # Data Controllerse
        self._devices: DeviceController = DeviceController(
            self
        )  # Devices contain all sensors
        self._fans: FanController = FanController(self)
        self._lights: LightController = LightController(self)
        self._locks: LockController = LockController(self)
        self._switches: SwitchController = SwitchController(self)
        self._valves: ValveController = ValveController(self)

    async def __aenter__(self) -> "HubspaceBridgeV1":
        """Return Context manager."""
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> bool | None:
        """Exit context manager."""
        await self.close()
        if exc_val:
            raise exc_val
        return exc_type

    @property
    def devices(self) -> DeviceController:
        return self._devices

    @property
    def events(self) -> EventStream:
        return self._events

    @property
    def fans(self) -> FanController:
        return self._fans

    @property
    def lights(self) -> LightController:
        return self._lights

    @property
    def locks(self) -> LockController:
        return self._locks

    @property
    def switches(self) -> SwitchController:
        return self._switches

    @property
    def valves(self) -> ValveController:
        return self._valves

    @property
    def _controllers(self) -> set:
        dev_controllers = {
            self._devices,
            self._fans,
            self._lights,
            self._locks,
            self._switches,
            self._valves,
        }
        return dev_controllers

    @property
    def controllers(self) -> set:
        initialized = set()
        for controller in self._controllers:
            if controller and controller.initialized:
                initialized.add(controller)
        return initialized

    @property
    def tracked_devices(self) -> set:
        if not self._known_devs:
            for controller in self.controllers:
                for device in controller.items:
                    self._known_devs.add(device.id)
        return self._known_devs

    @property
    def tracked_device_classes(self) -> set:
        dev_classes = set()
        for tracked_dc in self.device_classes.values():
            for cls in tracked_dc:
                dev_classes.add(cls)
        return dev_classes

    @property
    def device_classes(self) -> dict[str, list[str]]:
        if not self._known_dev_classes:
            for controller in self.controllers:
                self._known_dev_classes[controller.ITEM_CLS] = []
                for cls in controller.ITEM_TYPES:
                    self._known_dev_classes[controller.ITEM_CLS].append(cls.value)
        return self._known_dev_classes

    def add_device(self, device_id: str) -> None:
        self._known_devs.add(device_id)

    def remove_device(self, device_id: str) -> None:
        self._known_devs.remove(device_id)

    @property
    def account_id(self) -> str:
        """Get the account ID for the Hubspace account"""
        return self._account_id

    def set_polling_interval(self, polling_interval: int) -> None:
        self._events.polling_interval = polling_interval

    async def close(self) -> None:
        """Close connection and cleanup."""
        await self.events.stop()
        if self._close_session and self._web_session:
            await self._web_session.close()
        self.logger.info("Connection to bridge closed.")

    def subscribe(
        self,
        callback: EventCallBackType,
    ) -> Callable:
        """
        Subscribe to status changes for all resources.

        Returns:
            function to unsubscribe.
        """
        unsubscribes = [
            controller.subscribe(callback) for controller in self.controllers
        ]

        def unsubscribe():
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def get_account_id(self) -> str:
        """Lookup the account ID associated with the login"""
        self.logger.debug("Querying API for account id")
        headers = {"host": "api2.afero.net"}
        res = await self.request(
            "GET", v1_const.HUBSPACE_ACCOUNT_ID_URL, headers=headers
        )
        return (
            (await res.json()).get("accountAccess")[0].get("account").get("accountId")
        )

    async def initialize(self) -> None:
        """Query Hubspace API for all data"""
        if not self._account_id:
            self._account_id = await self.get_account_id()
        hs_data = await self.fetch_data()
        await asyncio.gather(
            *[
                controller.initialize(hs_data)
                for controller in self._controllers
                if not controller.initialized
            ]
        )
        await self._events.initialize()

    async def fetch_data(self) -> list[dict[Any, str]]:
        """Query the API"""
        self.logger.debug("Querying API for all data")
        headers = {
            "host": v1_const.HUBSPACE_DATA_HOST,
        }
        params = {"expansions": "state"}
        res = await self.request(
            "get",
            v1_const.HUBSPACE_DATA_URL.format(self.account_id),
            headers=headers,
            params=params,
        )
        return await res.json()

    @asynccontextmanager
    async def create_request(
        self, method: str, url: str, **kwargs
    ) -> Generator[aiohttp.ClientResponse, None, None]:
        """
        Make a request to any path with V2 request method (auth in header).

        Returns a generator with aiohttp ClientResponse.
        """
        if self._web_session is None:
            connector = aiohttp.TCPConnector(
                limit_per_host=3,
            )
            self._web_session = aiohttp.ClientSession(connector=connector)

        token = await self._auth.token(self._web_session)
        headers = get_headers(
            **{
                "authorization": f"Bearer {token}",
            }
        )
        headers.update(kwargs.get("headers", {}))
        kwargs["headers"] = headers
        kwargs["ssl"] = True
        async with self._web_session.request(method, url, **kwargs) as res:
            yield res

    async def request(self, method: str, url: str, **kwargs) -> aiohttp.ClientResponse:
        """Make request on the api and return response data."""
        retries = 0
        self.logger.info("Making request [%s] to %s with %s", method, url, kwargs)
        while retries < v1_const.MAX_RETRIES:
            retries += 1
            if retries > 1:
                retry_wait = 0.25 * retries
                await asyncio.sleep(retry_wait)
            async with self.create_request(method, url, **kwargs) as resp:
                # 503 means the service is temporarily unavailable, back off a bit.
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status in [429, 503]:
                    continue
                # 403 is bad auth
                elif resp.status == 403:
                    raise web_exceptions.HTTPForbidden()
                await resp.read()
                return resp
        raise ExceededMaximumRetries("Exceeded maximum number of retries")


def get_headers(**kwargs):
    headers = copy.copy(v1_const.DEFAULT_HEADERS)
    headers.update(kwargs)
    return headers



from dataclasses import dataclass, field

MAPPED_SENSORS = [
    "battery-level",
    "output-voltage-switch",
    "watts",
    "wifi-rssi",
]

BINARY_SENSORS = ["error"]


@dataclass
class HubspaceSensor:
    id: str
    owner: str
    value: str | int | float | None

    unit: str | None = field(default=None)
    instance: str | None = field(default=None)



import pytest

from aiohubspace.v1.controllers.device import DeviceController
from aiohubspace.v1.models.resource import DeviceInformation
from aiohubspace.v1.models.sensor import HubspaceSensor

from .. import utils

a21_light = utils.create_devices_from_data("light-a21.json")[0]
zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]
freezer = utils.create_devices_from_data("freezer.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = DeviceController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize_a21(mocked_controller):
    await mocked_controller.initialize_elem(a21_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == a21_light.id
    assert dev.available is True
    assert dev.device_information == DeviceInformation(
        device_class=a21_light.device_class,
        default_image=a21_light.default_image,
        default_name=a21_light.default_name,
        manufacturer=a21_light.manufacturerName,
        model=a21_light.model,
        name=a21_light.friendly_name,
        parent_id=a21_light.device_id,
    )
    assert dev.sensors == {
        "wifi-rssi": HubspaceSensor(
            id="wifi-rssi",
            owner="30a2df8c-109b-42c2-aed6-a6b30c565f8f",
            value=-50,
            instance=None,
            unit="dB",
        )
    }
    assert dev.binary_sensors == {}


@pytest.mark.asyncio
async def test_initialize_binary_sensors(mocked_controller):
    await mocked_controller.initialize_elem(freezer)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == freezer.id
    assert dev.available is True
    assert dev.device_information == DeviceInformation(
        device_class=freezer.device_class,
        default_image=freezer.default_image,
        default_name=freezer.default_name,
        manufacturer=freezer.manufacturerName,
        model=freezer.model,
        name=freezer.friendly_name,
        parent_id=freezer.device_id,
    )
    assert dev.sensors == {
        "wifi-rssi": HubspaceSensor(
            id="wifi-rssi",
            owner="596c120d-4e0d-4e33-ae9a-6330dcf2cbb5",
            value=-71,
            instance=None,
            unit="dB",
        )
    }
    assert dev.binary_sensors == {
        "error|freezer-high-temperature-alert": HubspaceSensor(
            id="error|freezer-high-temperature-alert",
            owner="596c120d-4e0d-4e33-ae9a-6330dcf2cbb5",
            value="normal",
            instance="freezer-high-temperature-alert",
        ),
        "error|fridge-high-temperature-alert": HubspaceSensor(
            id="error|fridge-high-temperature-alert",
            owner="596c120d-4e0d-4e33-ae9a-6330dcf2cbb5",
            value="alerting",
            instance="fridge-high-temperature-alert",
        ),
        "error|mcu-communication-failure": HubspaceSensor(
            id="error|mcu-communication-failure",
            owner="596c120d-4e0d-4e33-ae9a-6330dcf2cbb5",
            value="normal",
            instance="mcu-communication-failure",
        ),
        "error|temperature-sensor-failure": HubspaceSensor(
            id="error|temperature-sensor-failure",
            owner="596c120d-4e0d-4e33-ae9a-6330dcf2cbb5",
            value="normal",
            instance="temperature-sensor-failure",
        ),
    }


@pytest.mark.xfail(reason="Expecting raw HS data and given devices")
@pytest.mark.parametrize("file, expected_keys", [("light-a21.json", [a21_light.id])])
def test_get_filtered_devices(file, expected_keys, mocked_controller):
    data = utils.get_device_dump("light-a21.json")
    res = mocked_controller.get_filtered_devices(data)
    assert len(res) == len(expected_keys)
    for key in expected_keys:
        assert key in res



class HubspaceError(Exception):
    pass


class DeviceNotFound(HubspaceError):
    pass


class DeviceUpdateError(HubspaceError):
    pass


class ExceededMaximumRetries(HubspaceError):
    pass


class InvalidAuth(HubspaceError):
    pass


class InvalidResponse(HubspaceError):
    pass



from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes


@dataclass
class Lock:
    """Representation of a Hubspace Lock"""

    id: str  # ID used when interacting with Hubspace
    available: bool

    position: features.CurrentPositionFeature
    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)

    type: ResourceTypes = ResourceTypes.FAN

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            try:
                if function["functionInstance"]:
                    instances[function["functionClass"]] = function["functionInstance"]
            except KeyError:
                continue
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)


@dataclass
class LockPut:
    """States that can be updated for a Lock"""

    position: features.CurrentPositionFeature | None = None



from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes


@dataclass
class Fan:
    """Representation of a Hubspace Fan"""

    id: str  # ID used when interacting with Hubspace
    available: bool

    on: features.OnFeature
    speed: features.SpeedFeature
    direction: features.DirectionFeature
    preset: features.PresetFeature

    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)

    type: ResourceTypes = ResourceTypes.FAN

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            try:
                if function["functionInstance"]:
                    instances[function["functionClass"]] = function["functionInstance"]
            except KeyError:
                continue
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)

    @property
    def supports_direction(self):
        return self.direction is not None

    @property
    def supports_on(self):
        return self.on is not None

    @property
    def supports_presets(self):
        return self.preset is not None

    @property
    def supports_speed(self):
        return self.speed is not None

    @property
    def is_on(self) -> bool:
        """Return bool if fan is currently powered on."""
        if self.on is not None:
            return self.on.on
        return False

    @property
    def current_direction(self) -> bool:
        """Return if the direction is forward"""
        return self.direction.forward

    @property
    def current_speed(self) -> int:
        """Current speed of the fan, as a percentage"""
        return self.speed.speed

    @property
    def current_preset(self) -> str | None:
        """Current fan preset"""
        if self.preset.enabled:
            return self.preset.func_instance
        else:
            return None


@dataclass
class FanPut:
    """States that can be updated for a Fan"""

    on: features.OnFeature | None = None
    speed: features.SpeedFeature | None = None
    direction: features.DirectionFeature | None = None
    preset: features.PresetFeature | None = None



"""Test SwitchController"""

import pytest

from aiohubspace.v1.controllers.switch import SwitchController, features
from aiohubspace.v1.device import HubspaceState

from .. import utils

switch = utils.create_devices_from_data("switch-HPDA311CWB.json")[0]
transformer = utils.create_devices_from_data("transformer.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = SwitchController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(switch)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "feb5d9db-0562-478b-aaa0-00c889f0a758"
    assert dev.on == {
        None: features.OnFeature(on=False, func_class="power", func_instance=None),
    }


@pytest.mark.asyncio
async def test_initialize_multi(mocked_controller):
    await mocked_controller.initialize_elem(transformer)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "f9aa07e9-a4ce-46b4-b6bc-ad3bc070bc90"
    assert dev.on == {
        None: features.OnFeature(on=False, func_class="power", func_instance=None),
        "zone-1": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-1"
        ),
        "zone-2": features.OnFeature(
            on=True, func_class="toggle", func_instance="zone-2"
        ),
        "zone-3": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-3"
        ),
    }


@pytest.mark.asyncio
async def test_turn_on(mocked_controller):
    await mocked_controller.initialize_elem(switch)
    dev = mocked_controller.items[0]
    await mocked_controller.turn_on(switch.id)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == switch.id
    expected_states = [
        {
            "functionClass": "power",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "on",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)
    assert dev.on == {
        None: features.OnFeature(on=True, func_class="power", func_instance=None)
    }


@pytest.mark.asyncio
async def test_turn_on_multi(mocked_controller):
    await mocked_controller.initialize_elem(transformer)
    dev = mocked_controller.items[0]
    await mocked_controller.turn_on(transformer.id, instance="zone-1")
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == transformer.id
    expected_states = [
        {
            "functionClass": "toggle",
            "functionInstance": "zone-1",
            "lastUpdateTime": 12345,
            "value": "on",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)
    assert dev.on == {
        None: features.OnFeature(on=False, func_class="power", func_instance=None),
        "zone-1": features.OnFeature(
            on=True, func_class="toggle", func_instance="zone-1"
        ),
        "zone-2": features.OnFeature(
            on=True, func_class="toggle", func_instance="zone-2"
        ),
        "zone-3": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-3"
        ),
    }


@pytest.mark.asyncio
async def test_turn_off(mocked_controller):
    await mocked_controller.initialize_elem(switch)
    dev = mocked_controller.items[0]
    await mocked_controller.turn_off(switch.id)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == switch.id
    expected_states = [
        {
            "functionClass": "power",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "off",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)
    assert dev.on == {
        None: features.OnFeature(on=False, func_class="power", func_instance=None)
    }


@pytest.mark.asyncio
async def test_turn_off_multi(mocked_controller):
    await mocked_controller.initialize_elem(transformer)
    dev = mocked_controller.items[0]
    await mocked_controller.turn_off(transformer.id, instance="zone-2")
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == transformer.id
    expected_states = [
        {
            "functionClass": "toggle",
            "functionInstance": "zone-2",
            "lastUpdateTime": 12345,
            "value": "off",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)
    assert dev.on == {
        None: features.OnFeature(on=False, func_class="power", func_instance=None),
        "zone-1": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-1"
        ),
        "zone-2": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-2"
        ),
        "zone-3": features.OnFeature(
            on=False, func_class="toggle", func_instance="zone-3"
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller.initialize_elem(transformer)
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("transformer.json")[0]
    new_states = [
        HubspaceState(
            **{
                "functionClass": "toggle",
                "value": "on",
                "lastUpdateTime": 0,
                "functionInstance": "zone-1",
            }
        ),
        HubspaceState(
            **{
                "functionClass": "toggle",
                "value": "off",
                "lastUpdateTime": 0,
                "functionInstance": "zone-2",
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.on["zone-1"].on is True
    assert dev.on["zone-2"].on is False



"""Controller holding and managing Hubspace resources of type `valve`."""

from ..device import HubspaceDevice
from ..models import features, valve
from ..models.resource import DeviceInformation, ResourceTypes
from .base import BaseResourcesController


class ValveController(BaseResourcesController[valve.Valve]):
    """Controller holding and managing Hubspace resources of type `valve`.

    A valve can have one or more toggleable elements. They are controlled
    by their functionInstance.
    """

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.WATER_TIMER]
    ITEM_CLS = valve.Valve
    ITEM_MAPPING = {}

    async def turn_on(self, device_id: str, instance: str | None = None) -> None:
        """Open the valve"""
        await self.set_state(device_id, valve_open=True, instance=instance)

    async def turn_off(self, device_id: str, instance: str | None = None) -> None:
        """Close the valve"""
        await self.set_state(device_id, valve_open=False, instance=instance)

    async def initialize_elem(self, hs_device: HubspaceDevice) -> None:
        """Initialize the element"""
        self._logger.info("Initializing %s", hs_device.id)
        available: bool = False
        valve_open: dict[str, features.OpenFeature] = {}
        for state in hs_device.states:
            if state.functionClass in ["power", "toggle"]:
                valve_open[state.functionInstance] = features.OpenFeature(
                    open=state.value == "on",
                    func_class=state.functionClass,
                    func_instance=state.functionInstance,
                )
            elif state.functionClass == "available":
                available = state.value

        self._items[hs_device.id] = valve.Valve(
            hs_device.functions,
            id=hs_device.id,
            available=available,
            device_information=DeviceInformation(
                device_class=hs_device.device_class,
                default_image=hs_device.default_image,
                default_name=hs_device.default_name,
                manufacturer=hs_device.manufacturerName,
                model=hs_device.model,
                name=hs_device.friendly_name,
                parent_id=hs_device.device_id,
            ),
            open=valve_open,
        )

    async def update_elem(self, hs_device: HubspaceDevice) -> None:
        cur_item = self.get_device(hs_device.id)
        for state in hs_device.states:
            if state.functionClass in ["power", "toggle"]:
                cur_item.open[state.functionInstance].open = state.value == "on"
            elif state.functionClass == "available":
                cur_item.available = state.value

    async def set_state(
        self,
        device_id: str,
        valve_open: bool | None = None,
        instance: str | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = valve.ValvePut()
        if valve_open is not None:
            dev = self.get_device(device_id)
            try:
                update_obj.open = features.OpenFeature(
                    open=valve_open,
                    func_class=dev.open[instance].func_class,
                    func_instance=instance,
                )
            except KeyError:
                self._logger.info("Unable to find instance %s", instance)
        await self.update(device_id, update_obj)



"""Handle connecting to the HUE Eventstream and distribute events."""

import asyncio
from asyncio.coroutines import iscoroutinefunction
from collections.abc import Callable
from enum import Enum
from types import NoneType
from typing import TYPE_CHECKING, NotRequired, TypedDict

from aiohttp.client_exceptions import ClientError

from ..device import HubspaceDevice, get_hs_device

if TYPE_CHECKING:
    from .. import HubspaceBridgeV1


class EventStreamStatus(Enum):
    """Status options of EventStream."""

    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class EventType(Enum):
    """Enum with possible Events."""

    # resource events match those emitted by Hue eventstream
    RESOURCE_ADDED = "add"
    RESOURCE_UPDATED = "update"
    RESOURCE_DELETED = "delete"
    # connection events emitted by (this) events controller
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    RECONNECTED = "reconnected"


class HubspaceEvent(TypedDict):
    """Hue Event message as emitted by the EventStream."""

    type: EventType  # = EventType (add, update, delete)
    device_id: str  # ID for interacting with the device
    device: NotRequired[HubspaceDevice]  # Hubspace Device


EventCallBackType = Callable[[EventType, dict | None], None]
EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
    "tuple[ResourceTypes] | None",
]


class EventStream:
    """Holds the connection to the HUE Clip EventStream."""

    def __init__(self, bridge: "HubspaceBridgeV1", polling_interval: int) -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._listeners = set()
        self._event_queue = asyncio.Queue()
        self._status = EventStreamStatus.DISCONNECTED
        self._bg_tasks: list[asyncio.Task] = []
        self._subscribers: list[EventSubscriptionType] = []
        self._logger = bridge.logger.getChild("events")
        self._polling_interval: int = polling_interval

    @property
    def connected(self) -> bool:
        """Return bool if we're connected."""
        return self._status == EventStreamStatus.CONNECTED

    @property
    def status(self) -> EventStreamStatus:
        """Return connection status."""
        return self._status

    @property
    def polling_interval(self) -> int:
        return self._polling_interval

    @polling_interval.setter
    def polling_interval(self, polling_interval: int) -> None:
        self._polling_interval = polling_interval

    async def initialize(self) -> None:
        """Start the polling processes"""
        assert len(self._bg_tasks) == 0
        self._bg_tasks.append(asyncio.create_task(self.__event_reader()))
        self._bg_tasks.append(asyncio.create_task(self.__event_processor()))

    async def stop(self) -> None:
        """Stop listening for events."""
        for task in self._bg_tasks:
            task.cancel()
        self._bg_tasks = []

    def subscribe(
        self,
        callback: EventCallBackType,
        event_filter: EventType | tuple[EventType] | None = None,
        resource_filter: tuple[str] | None = None,
    ) -> Callable:
        """
        Subscribe to events emitted

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `event_filter` - Optionally provide an EventType as filter.
            - `resource_filter` - Optionally provide a ResourceType as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, NoneType | tuple):
            event_filter = (event_filter,)
        if not isinstance(resource_filter, NoneType | tuple):
            resource_filter = (resource_filter,)
        subscription = (callback, event_filter, resource_filter)

        def unsubscribe():
            self._subscribers.remove(subscription)

        self._subscribers.append(subscription)
        return unsubscribe

    def emit(self, event_type: EventType, data: HubspaceEvent = None) -> None:
        """Emit event to all listeners."""
        for callback, event_filter, resource_filter in self._subscribers:
            try:
                if event_filter is not None and event_type not in event_filter:
                    continue
                if data is not None:
                    if resource_filter is not None:
                        if (
                            "device" in data
                            and data["device"]
                            and not any(
                                data["device"].device_class == res_filter
                                for res_filter in resource_filter
                            )
                        ):
                            continue
                if iscoroutinefunction(callback):
                    asyncio.create_task(callback(event_type, data))
                else:
                    callback(event_type, data)
            except Exception:
                self._logger.exception("Unhandled exception. Please open a bug report")

    async def __event_reader(self) -> None:
        """Poll the current states"""
        self._status = EventStreamStatus.CONNECTING
        while True:
            processed_ids = []
            skipped_ids = []
            try:
                data = await self._bridge.fetch_data()
                for dev in data:
                    hs_dev = get_hs_device(dev)
                    event_type = EventType.RESOURCE_UPDATED
                    if hs_dev.id not in self._bridge.tracked_devices:
                        event_type = EventType.RESOURCE_ADDED
                        self._bridge.add_device(hs_dev.id)
                    if hs_dev.device_class not in self._bridge.tracked_device_classes:
                        skipped_ids.append(hs_dev.id)
                        continue
                    self._event_queue.put_nowait(
                        HubspaceEvent(
                            type=event_type,
                            device_id=hs_dev.id,
                            device=hs_dev,
                        )
                    )
                    processed_ids.append(hs_dev.id)
            except (ClientError, asyncio.TimeoutError) as err:
                # Auto-retry will take care of the issue
                self._logger.warning(err)
            except Exception as err:
                self._logger.warning(err)
                raise err
            else:
                # Connection was successful. Find missing devices
                for dev_id in self._bridge.tracked_devices:
                    if dev_id not in processed_ids + skipped_ids:
                        self._event_queue.put_nowait(
                            HubspaceEvent(
                                type=EventType.RESOURCE_DELETED, device_id=dev_id
                            )
                        )
                        self._bridge.remove_device(dev_id)
            self._status = EventStreamStatus.DISCONNECTED
            await asyncio.sleep(self._polling_interval)

    async def __event_processor(self) -> None:
        """Process the hubspace devices"""
        while True:
            try:
                event: HubspaceEvent = await self._event_queue.get()
                self.emit(event["type"], event)
            except Exception:
                self._logger.exception("Unhandled exception. Please open a bug report")



__all__ = [
    "HubspaceDevice",
    "HubspaceState",
    "get_hs_device",
]
import logging
from dataclasses import dataclass, field
from typing import Any, Optional

logger = logging.getLogger(__name__)


@dataclass
class HubspaceState:
    """State of a given function

    :param functionClass: Function class for the state (ie, power)
    :param value: Value to set for the function_class
    :param lastUpdateTime: Last time the state was updated (in epoch ms).
    :param functionInstance: Additional information about the function (ie, light-power).
        Default: None
    """

    functionClass: str
    value: Any
    lastUpdateTime: Optional[int] = None
    functionInstance: Optional[str] = None


@dataclass
class HubspaceDevice:
    id: str
    device_id: str
    model: str
    device_class: str
    default_name: str
    default_image: str
    friendly_name: str
    functions: list[dict] = field(default=list)
    states: list[HubspaceState] = field(default=list)
    children: list[str] = field(default=list)
    manufacturerName: Optional[str] = field(default=None)

    def __hash__(self):
        return hash((self.id, self.friendly_name))

    def __post_init__(self):
        if not self.model and self.default_image == "ceiling-fan-snyder-park-icon":
            self.model = "DriskolFan"
        elif not self.model and self.default_image == "ceiling-fan-vinings-icon":
            self.model = "VinwoodFan"
        elif (
            self.device_class == "fan"
            and self.model == "TBD"
            and self.default_image == "ceiling-fan-chandra-icon"
        ):
            self.model = "ZandraFan"
        elif (
            self.model == "TBD"
            and self.default_image == "ceiling-fan-ac-cct-dardanus-icon"
        ):
            self.model = "NevaliFan"
        elif (
            self.device_class == "fan"
            and not self.model
            and self.default_image == "ceiling-fan-slender-icon"
        ):
            self.model = "TagerFan"
        elif self.model == "Smart Stake Timer":
            self.model = "YardStake"
        elif self.default_image == "a19-e26-color-cct-60w-smd-frosted-icon":
            self.model = "12A19060WRGBWH2"
        elif (
            self.device_class == "switch" and self.default_image == "slide-dimmer-icon"
        ):
            self.model = "HPDA110NWBP"
        # Dimmer Switch fix - A switch cannot dim, but a light can
        elif self.device_class == "switch" and any(
            [state.functionClass == "brightness" for state in self.states]
        ):
            self.device_class = "light"
        elif (
            self.device_class == "switch"
            and self.default_image == "smart-switch-icon"
            and self.model == "TBD"
        ):
            self.model = "HPSA11CWB"


def get_hs_device(hs_device: dict[str, Any]) -> HubspaceDevice:
    """Convert the Hubspace device definition into a HubspaceDevice"""
    description = hs_device.get("description", {})
    device = description.get("device", {})
    processed_states: list[HubspaceState] = []
    for state in hs_device.get("state", {}).get("values", []):
        processed_states.append(
            HubspaceState(
                functionClass=state.get("functionClass"),
                value=state.get("value"),
                lastUpdateTime=state.get("lastUpdateTime"),
                functionInstance=state.get("functionInstance"),
            )
        )
    dev_dict = {
        "id": hs_device.get("id"),
        "device_id": hs_device.get("deviceId"),
        "model": device.get("model"),
        "device_class": device.get("deviceClass"),
        "default_name": device.get("defaultName"),
        "default_image": description.get("defaultImage"),
        "friendly_name": hs_device.get("friendlyName"),
        "functions": description.get("functions", []),
        "states": processed_states,
        "children": hs_device.get("children", []),
        "manufacturerName": device.get("manufacturerName"),
    }
    return HubspaceDevice(**dev_dict)


def get_function_from_device(
    hs_device: HubspaceDevice, function_class: str, function_instance
) -> dict:
    for func in hs_device.functions:
        if func.get("functionClass") != function_class:
            continue
        if func.get("functionInstance") != function_instance:
            continue
        return func
    return None
